import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();
        scene.background = new THREE.Color(0xf0f0f0);

        let renderer = new THREE.WebGLRenderer( { antialias: true });
        this.domWebGL.appendChild(renderer.domElement);

        this._stats = new Stats();
        this.domWebGL.appendChild(this._stats.dom);

        window.onresize = this.resize.bind(this);

        this.renderer = renderer;
        this.scene = scene;
        
        this._setupCamera();
        this._setupModel();
        this._setupLights();
     
        this.resize();
    }

    _setupModel() {
        const group = new THREE.Group();
        group.position.y = 50;
        this.scene.add(group);

        const loader = new THREE.TextureLoader();
        const texture = loader.load('textures/uv_grid_opengl.jpg');
        texture.wrapS = texture.wrapT = THREE.RepeatWrapping;
        texture.repeat.set(0.008, 0.008);
        
        this._texture = texture;
        this._group = group;

        // California
        const californiaPts = [];

        californiaPts.push(new THREE.Vector2(610, 320));
        californiaPts.push(new THREE.Vector2(450, 300));
        californiaPts.push(new THREE.Vector2(392, 392));
        californiaPts.push(new THREE.Vector2(266, 438));
        californiaPts.push(new THREE.Vector2(190, 570));
        californiaPts.push(new THREE.Vector2(190, 600));
        californiaPts.push(new THREE.Vector2(160, 620));
        californiaPts.push(new THREE.Vector2(160, 650));
        californiaPts.push(new THREE.Vector2(180, 640));
        californiaPts.push(new THREE.Vector2(165, 680));
        californiaPts.push(new THREE.Vector2(150, 670));
        californiaPts.push(new THREE.Vector2(90, 737));
        californiaPts.push(new THREE.Vector2(80, 795));
        californiaPts.push(new THREE.Vector2(50, 835));
        californiaPts.push(new THREE.Vector2(64, 870));
        californiaPts.push(new THREE.Vector2(60, 945));
        californiaPts.push(new THREE.Vector2(300, 945));
        californiaPts.push(new THREE.Vector2(300, 743));
        californiaPts.push(new THREE.Vector2(600, 473));
        californiaPts.push(new THREE.Vector2(626, 425));
        californiaPts.push(new THREE.Vector2(600, 370));
        californiaPts.push(new THREE.Vector2(610, 320));

        for (let i = 0; i < californiaPts.length; i++) californiaPts[i].multiplyScalar(0.25);
        const californiaShape = new THREE.Shape(californiaPts);

        // Triangle
        const triangleShape = new THREE.Shape()
            .moveTo(80, 20)
            .lineTo(40, 80)
            .lineTo(120, 80)
            .lineTo(80, 20); // close path 

        // Heart
        const x = 0, y = 0;
        const heartShape = new THREE.Shape() // From http://blog.burlock.org/html5/130-paths
            .moveTo(x + 25, y + 25)
            .bezierCurveTo(x + 25, y + 25, x + 20, y, x, y)
            .bezierCurveTo(x - 30, y, x - 30, y + 35, x - 30, y + 35)
            .bezierCurveTo(x - 30, y + 55, x - 10, y + 77, x + 25, y + 95)
            .bezierCurveTo(x + 60, y + 77, x + 80, y + 55, x + 80, y + 35)
            .bezierCurveTo(x + 80, y + 35, x + 80, y, x + 50, y)
            .bezierCurveTo(x + 35, y, x + 25, y + 25, x + 25, y + 25);

        // Square
        const sqLength = 80;
        const squareShape = new THREE.Shape()
            .moveTo(0, 0)
            .lineTo(0, sqLength)
            .lineTo(sqLength, sqLength)
            .lineTo(sqLength, 0)
            .lineTo(0, 0);

        // Rounded rectangle
        const roundedRectShape = new THREE.Shape();
        (function roundedRect(ctx, x, y, width, height, radius) {
            ctx.moveTo(x, y + radius);
            ctx.lineTo(x, y + height - radius);
            ctx.quadraticCurveTo(x, y + height, x + radius, y + height);
            ctx.lineTo(x + width - radius, y + height);
            ctx.quadraticCurveTo(x + width, y + height, x + width, y + height - radius);
            ctx.lineTo(x + width, y + radius);
            ctx.quadraticCurveTo(x + width, y, x + width - radius, y);
            ctx.lineTo(x + radius, y);
            ctx.quadraticCurveTo(x, y, x, y + radius);

        })(roundedRectShape, 0, 0, 50, 50, 20);

        // Track
        const trackShape = new THREE.Shape()
            .moveTo(40, 40)
            .lineTo(40, 160)
            .absarc(60, 160, 20, Math.PI, 0, true)
            .lineTo(80, 40)
            .absarc(60, 40, 20, 2 * Math.PI, Math.PI, true);

        // Circle
        const circleRadius = 40;
        const circleShape = new THREE.Shape()
            .moveTo(0, circleRadius)
            .quadraticCurveTo(circleRadius, circleRadius, circleRadius, 0)
            .quadraticCurveTo(circleRadius, - circleRadius, 0, - circleRadius)
            .quadraticCurveTo(- circleRadius, - circleRadius, - circleRadius, 0)
            .quadraticCurveTo(- circleRadius, circleRadius, 0, circleRadius);

        // Fish
        const fishShape = new THREE.Shape()
            .moveTo(x, y)
            .quadraticCurveTo(x + 50, y - 80, x + 90, y - 10)
            .quadraticCurveTo(x + 100, y - 10, x + 115, y - 40)
            .quadraticCurveTo(x + 115, y, x + 115, y + 40)
            .quadraticCurveTo(x + 100, y + 10, x + 90, y + 10)
            .quadraticCurveTo(x + 50, y + 80, x, y);

        // Arc circle
        const arcShape = new THREE.Shape()
            .moveTo(50, 10)
            .absarc(10, 10, 40, 0, Math.PI * 2, false);

        const holePath = new THREE.Path()
            .moveTo(20, 10)
            .absarc(10, 10, 10, 0, Math.PI * 2, true);

        arcShape.holes.push(holePath);

        // Smiley
        const smileyShape = new THREE.Shape()
            .moveTo(80, 40)
            .absarc(40, 40, 40, 0, Math.PI * 2, false);

        const smileyEye1Path = new THREE.Path()
            .moveTo(35, 20)
            .absellipse(25, 20, 10, 10, 0, Math.PI * 2, true);

        const smileyEye2Path = new THREE.Path()
            .moveTo(65, 20)
            .absarc(55, 20, 10, 0, Math.PI * 2, true);

        const smileyMouthPath = new THREE.Path()
            .moveTo(20, 40)
            .quadraticCurveTo(40, 60, 60, 40)
            .bezierCurveTo(70, 45, 70, 50, 60, 60)
            .quadraticCurveTo(40, 80, 20, 60)
            .quadraticCurveTo(5, 50, 20, 40);

        smileyShape.holes.push(smileyEye1Path);
        smileyShape.holes.push(smileyEye2Path);
        smileyShape.holes.push(smileyMouthPath);

        // Spline shape
        const splinepts = [];
        splinepts.push(new THREE.Vector2(70, 20));
        splinepts.push(new THREE.Vector2(80, 90));
        splinepts.push(new THREE.Vector2(- 30, 70));
        splinepts.push(new THREE.Vector2(0, 0));

        const splineShape = new THREE.Shape()
            .moveTo(0, 0)
            .splineThru(splinepts);

        const extrudeSettings = { depth: 8, bevelEnabled: true, bevelSegments: 2, steps: 2, bevelSize: 1, bevelThickness: 1};

        this._addShape(californiaShape, extrudeSettings, 0xf08000, -300, -100, 0, 0, 0, 0, 1);
        this._addShape( triangleShape, extrudeSettings, 0x8080f0, - 180, 0, 0, 0, 0, 0, 1 );
        this._addShape( roundedRectShape, extrudeSettings, 0x008000, - 150, 150, 0, 0, 0, 0, 1 );
        this._addShape( trackShape, extrudeSettings, 0x008080, 200, - 100, 0, 0, 0, 0, 1 );
        this._addShape( squareShape, extrudeSettings, 0x0040f0, 150, 100, 0, 0, 0, 0, 1 );
        this._addShape( heartShape, extrudeSettings, 0xf00000, 60, 100, 0, 0, 0, Math.PI, 1 );
        this._addShape( circleShape, extrudeSettings, 0x00f000, 120, 250, 0, 0, 0, 0, 1 );
        this._addShape( fishShape, extrudeSettings, 0x404040, - 60, 200, 0, 0, 0, 0, 1 );
        this._addShape( smileyShape, extrudeSettings, 0xf000f0, - 200, 250, 0, 0, 0, Math.PI, 1 );
        this._addShape(arcShape, extrudeSettings, 0x804000, 150, 0, 0, 0, 0, 0, 1);
        this._addShape(splineShape, extrudeSettings, 0x808080, - 50, - 100, 0, 0, 0, 0, 1);

        this._addLineShape(arcShape.holes[0], 0x804000, 150, 0, 0, 0, 0, 0, 1);

        for (let i = 0; i < smileyShape.holes.length; i += 1) {
            this._addLineShape(smileyShape.holes[i], 0xf000f0, - 200, 250, 0, 0, 0, Math.PI, 1);
        }
    }

    _addShape(shape, extrudeSettings, color, x, y, z, rx, ry, rz, s) {
        // Flat Shape with Texture
        let geometry = new THREE.ShapeGeometry(shape);
        let mesh = new THREE.Mesh(geometry, new THREE.MeshPhongMaterial({
            side: THREE.DoubleSide, map: this._texture
        }));
        mesh.position.set(x, y, z-175);
        mesh.rotation.set(rx, ry, rz);
        mesh.scale.set(s, s, s);
        this._group.add(mesh);

        // Flat Shape
        geometry = new THREE.ShapeGeometry(shape);
        mesh = new THREE.Mesh(geometry, new THREE.MeshPhongMaterial({
            color: color, side: THREE.DoubleSide
        }));
        mesh.position.set(x, y, z-125);
        mesh.rotation.set(rx, ry, rz);
        mesh.scale.set(s, s, s);
        this._group.add(mesh);

        // Extruded Shape
        geometry = new THREE.ExtrudeGeometry(shape, extrudeSettings);
        mesh = new THREE.Mesh(geometry, new THREE.MeshPhongMaterial({ color: color }));
        mesh.position.set(x, y, z-75);
        mesh.rotation.set(rx, ry, rz);
        mesh.scale.set(s, s, s);
        this._group.add(mesh);

        // Line
        this._addLineShape(shape, color, x, y, z, rx, ry, rz, s);
    }

    _addLineShape(shape, color, x, y, z, rx, ry, rz, s) {
        shape.autoClose = true;
        const points = shape.getPoints();
        const spacedPoints = shape.getSpacedPoints(16); // 16개
        const geometryPoints = new THREE.BufferGeometry().setFromPoints(points);
        const geometrySpacedPoints = new THREE.BufferGeometry().setFromPoints(spacedPoints);

        // Solid Line
        let line = new THREE.Line(geometryPoints, new THREE.LineBasicMaterial({color: color}));
        line.position.set(x, y, z-25);
        line.rotation.set(rx, ry, rz);
        line.scale.set(s, s, s);
        this._group.add(line);

        // Line from equidistance sampled points
        line = new THREE.Line(geometrySpacedPoints, new THREE.LineBasicMaterial({color: color}));
        line.position.set(x, y, z+25);
        line.rotation.set(rx, ry, rz);
        line.scale.set(s, s, s);
        this._group.add(line);

        // Vertices from real points
        let particles = new THREE.Points(geometryPoints, new THREE.PointsMaterial({color:color, size:4}));
        particles.position.set(x, y, z+75);
        particles.rotation.set(rx, ry, rz);
        particles.scale.set(s, s, s);
        this._group.add(particles);

        // Vertices from equidistance sampled points
        particles = new THREE.Points(geometrySpacedPoints, new THREE.PointsMaterial({ color: color, size: 4}));
        particles.position.set(x, y, z+125);
        particles.rotation.set(rx, ry, rz);
        particles.scale.set(s, s, s);
        this._group.add(particles);
    }

    _setupLights() {
        const pointLight = new THREE.PointLight(0xffffff, 0.8);
        this._camera.add(pointLight);

    }

    _setupCamera() {
        const aspect = window.innerWidth / window.innerHeight;
   
        const camera = new THREE.PerspectiveCamera(50, aspect, 1, 1000);
        camera.position.set(0, 160, 700);
        this.scene.add(camera);
        this._camera = camera;
    }

    update(time) {
        this._group.rotation.y = time * 0.001;
        this._stats.update();
    }

    render(time) {
        requestAnimationFrame(this.render.bind(this));

        this.update(time);
        
        this.renderer.render(this.scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();
    }
}

window.onload = function () {
    (new App()).render();
}